/***************************************************************************
                          CNeuralNet.cpp  -  description
                             -------------------
    begin                : Sat Dec 1 2001
    copyright            : (C) 2001 by Gene Ruebsamen
    email                : gene@erachampion.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <cassert>
#include <cstdlib>
#include <cmath>

using std::cout;
using std::endl;

#include "ann.h"
#include "net.h"
#include "constants.h"

ANN::ANN(int in, int hide, int out, int recur) : neuralNet(in,hide,out,recur)
{
	// calculate size of Weight Array
	int length = neuralNet.getSize();
	weightPtr = new float[length];			// allocate storage
	assert(weightPtr != 0);

	for (int i=0;i<length;i++) {
		if (rand()%1000 > 500)
			weightPtr[i] =  (float)((double)rand() / (double)(RAND_MAX+1));	// random number between [0,1]
		else
			weightPtr[i] = - (float)((double)rand() / (double)(RAND_MAX+1));	// random number between [-1,0]
	}
	neuralNet.initWeights(weightPtr);		// initialize Neural Network
}

ANN::ANN(const ANN &copy) : neuralNet(copy.neuralNet)			// copy constructor
{
	delete [] weightPtr;
	setWeights(copy.weightPtr);
	fit_score = copy.fit_score;
}

ANN::~ANN()
{
	delete [] weightPtr;
}

void ANN::setFitness(float fit)
{
	fit_score = fit;
}

ostream &operator<<(ostream &output, const ANN &right)
{
	output << "[";
	for (int i=0;i<right.neuralNet.getSize();i++)
	{
		if (i < right.neuralNet.getSize()-1)
			output << right.weightPtr[i] << " ";
	}
	output << "]";

	return output;
}

const ANN &ANN::operator=(const ANN &right)
{
	if (&right != this)			// avoid self-assignment
	{
		delete [] weightPtr;
		neuralNet = right.neuralNet;
		setWeights(right.weightPtr);
		fit_score = right.fit_score;
	}
	else
		cout << "Attempted assignment of an ANN to itself!" << endl;
	return *this;
}

bool ANN::operator==(const ANN &right) const
{
	if (neuralNet.getInputSize() != right.neuralNet.getInputSize())
		return false;
	if (neuralNet.getHiddenSize() != right.neuralNet.getHiddenSize())
		return false;
	if (neuralNet.getOutputSize() != right.neuralNet.getOutputSize())
		return false;
	if (neuralNet.getRecurrentSize() != right.neuralNet.getRecurrentSize())
		return false;
	else {
		for (int i=0;i<neuralNet.getSize();i++)
			if (weightPtr[i] != right.weightPtr[i])
				return false;
	}
	return true;
}

int ANN::getLength(void) const
{
	return neuralNet.getSize();
}

//--------
// Main Reproduction algorithms for the two ANN's
//--------
const ANN *ANN::operator+(const ANN &right)
{
	ANN *tempANN = new ANN(neuralNet.getInputSize(),neuralNet.getHiddenSize(),
												 neuralNet.getOutputSize(),neuralNet.getRecurrentSize());

	int length = neuralNet.getSize();
	float tempWeight[length];

  // Chance of Crossover
	if ((((float)(rand() %1000))/1000.0) <= CROSSOVER)
	{
	 	int crosspoint = (rand()%length);

    // Determine Crossover Point
 		for (int i=0;i<length;i++)
		{
			if (i <= crosspoint)
				tempWeight[i] = weightPtr[i];
			else
				tempWeight[i] = right.weightPtr[i];
		}
	}
	else // just copy
	{
		if (rand()%1000 > 500) {
			for (int i=0;i<length;i++)
				tempWeight[i] = weightPtr[i];
		}
		else {
			for (int i=0;i<length;i++)
				tempWeight[i] = right.weightPtr[i];
		}
	}

	// Chance of Mutation
	for (int i=0;i<length;i++)
	{
		if ((((float)(rand()%1000))/1000.0) <= MUTATION)
		{
			// Gaussian random perturbation utilizing Box-Mueller Transform
	 		float x1, x2, w, y1, y2;

 	    do {
 		   	x1 = 2.0 * ((rand()%10000)/10000.0) - 1.0;
 	     	x2 = 2.0 * ((rand()%10000)/10000.0) - 1.0;
 	     	w = x1 * x1 + x2 * x2;
   	 	} while ( w >= 1.0 );

   	 	w = sqrt( (-2.0 * log( w ) ) / w );
   	 	y1 = x1 * w;
    	y2 = x2 * w;

			tempWeight[i] = tempWeight[i] + sqrt(0.15) * y1;
		}
	}
	delete [] tempANN->weightPtr;         // delete memory
	tempANN->setWeights(tempWeight);	    // allocate memory
	return tempANN;
}

void ANN::resetANN()
{
	fit_score=false;
}

// Return Fitness Score
float ANN::getFitness(void)
{
	return fit_score;
}

// pass in input to ANN and get output back
const float * const ANN::getOutput(const float * const input)
{
	return neuralNet.getOutput(input);
}

void ANN::setWeights(float *weights)
{
	int size = neuralNet.getSize();
	weightPtr = new float[size];			  // allocate storage
	assert(weightPtr != 0);
	for (int i=0;i<size;i++)
		weightPtr[i] = weights[i];        // copy chromosome
	neuralNet.initWeights(weightPtr);		// initialize Neural Network
}
